﻿Imports System.IO

Public Class Form1

  ' Grid is a byte array with the following properties
  ' Bit 0 is used to hold the next generation
  ' Bit 1 is the current generation
  ' Bits 2-7 are the previous generations

  Shadows scale As Integer
  Dim speed As String
  Dim mode As String = "Rectangular"

  Private Sub Form1_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
    RadioButton1.Checked = True
    RadioButton6.Checked = True
    RadioButton9.Checked = True
    NumericUpDown1.Value = 3
    NumericUpDown2.Value = 3
    NumericUpDown3.Value = 2
    NumericUpDown4.Value = 3
    Label5.CheckForIllegalCrossThreadCalls = False
    Label6.CheckForIllegalCrossThreadCalls = False
    Ngens = 0 : Ncells = 0
    UpdateStats()
    scale = 20
    speed = "slow"
    Xmin = 480 : Xmax = 520 : Ymin = 480 : Ymax = 520
    ToDo = 0
    Buffer(0) = 1 << 31 : BufferPointer = 0
    LoadList()
    Me.BackgroundWorker1.RunWorkerAsync()
  End Sub

  Private Sub LoadList()
    Dim CurrentDir As String = System.Environment.CurrentDirectory.ToString
    Dim filename As String = CurrentDir & "\List.txt"
    Dim filestream As New FileStream(filename, FileMode.Open, FileAccess.Read)
    Dim streamreader As New StreamReader(filestream)
    Nlist = 0
    Try
      While Not streamreader.EndOfStream
        List(Nlist) = streamreader.ReadLine
        ComboBox1.Items.Add(GetName(List(Nlist)))
        Nlist += 1
      End While
      streamreader.Close()
    Catch ex As IOException
      MsgBox(ex.Message)
    End Try
    ComboBox1.Text = GetName(List(0))
  End Sub

  Private Sub SaveList()
    Dim CurrentDir As String = System.Environment.CurrentDirectory.ToString
    Dim filename As String = CurrentDir & "\List.txt"
    Dim filestream As New FileStream(filename, FileMode.Create, FileAccess.Write)
    Dim streamwriter As New StreamWriter(filestream)
    Try
      For i As Integer = 0 To Nlist - 1
        streamwriter.WriteLine(List(i))
      Next
      streamwriter.Close()
    Catch ex As IOException
      MsgBox(ex.Message)
    End Try
  End Sub

  Private Function GetName(ByVal t As String) As String
    Return t.Substring(0, t.IndexOf(":"))
  End Function

  Private Function GetData(ByVal t As String) As String
    Return t.Substring(t.IndexOf(":") + 1)
  End Function

  Private Sub PictureBox1_MouseClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles PictureBox1.MouseClick
    Dim i As Integer
    Dim j As Integer
    Select Case mode
      Case "Hexagonal"
        j = Div(e.Y - 400 - scale / 2, scale * 1.732 / 2)
        i = Div(e.X - 400 - scale / 2 + j * scale / 2, scale) + 500
        j += 500
      Case "Rectangular"
        i = Div(e.X - 400 - scale / 2, scale) + 500
        j = Div(e.Y - 400 - scale / 2, scale) + 500
    End Select
    If e.Button = Windows.Forms.MouseButtons.Left Then Paint(i, j, 2) Else Paint(i, j, 0)
    UpdateLimits(i, j)
    PictureBox1.Invalidate()
  End Sub

  Private Shadows Sub Paint(ByVal i, ByVal j, ByVal f)
    Dim n As Integer = ComboBox1.SelectedIndex
    If n < 0 Then Beep() : Exit Sub
    Dim x, y As Integer
    Dim t As String = List(n)
    t = GetData(t)
    While t <> ""
      x = Val(t)
      t = t.Substring(t.IndexOf(",") + 1)
      y = Val(t)
      t = t.Substring(t.IndexOf(",") + 1)
      Grid(i + x, j + y) = f
    End While
  End Sub

  Private Sub UpdateLimits(ByVal i, ByVal j)
    If i < Xmin Then Xmin = i
    If i > Xmax Then Xmax = i
    If j < Ymin Then Ymin = j
    If j > Ymax Then Ymax = j
  End Sub

  Private Function Div(ByVal a, ByVal b) As Integer
    Return CInt((a + b * 100000) / b) - 100000
  End Function

  Private Sub PictureBox1_Paint(ByVal sender As Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles PictureBox1.Paint
    Dim g As Graphics = e.Graphics
    Dim x, y As Single
    Select Case mode
      Case "Rectangular"
        g.FillEllipse(Brushes.White, 400, 400, scale, scale)
        For i As Integer = Xmin To Xmax
          For j As Integer = Ymin To Ymax
            x = i - 500 : y = j - 500 : RectCoords(x, y)
            If Grid(i, j) And 2 Then
              g.FillEllipse(Brushes.Black, x, y, scale, scale)
            ElseIf Grid(i, j) And 4 Then
              g.FillEllipse(Brushes.Khaki, x, y, scale, scale)
            End If
          Next
        Next
        Dim n As Integer = 400 \ scale + 1
        If n < 100 Then
          For i As Integer = 400 - n * scale To 400 + n * scale Step scale
            g.DrawLine(Pens.LightGray, i, 0, i, 800)
            g.DrawLine(Pens.LightGray, 0, i, 800, i)
          Next
        End If
      Case "Hexagonal"
        g.FillEllipse(Brushes.White, 400, 400, scale, scale)
        For i As Integer = Xmin To Xmax
          For j As Integer = Ymin To Ymax
            x = i - 500 : y = j - 500 : HexCoords(x, y)
            If Grid(i, j) And 2 Then
              g.FillEllipse(Brushes.Black, x, y, scale, scale)
            ElseIf Grid(i, j) And 4 Then
              g.FillEllipse(Brushes.Khaki, x, y, scale, scale)
            End If
          Next
        Next
    End Select
  End Sub

  Private Sub RectCoords(ByRef x As Single, ByRef y As Single)
    ' x an y are supplied as integer grid coords and returned as fp cartesian coords
    x = x * scale + 400
    y = y * scale + 400
  End Sub

  Private Sub HexCoords(ByRef x As Single, ByRef y As Single)
    ' x an y are supplied as integer grid coords and returned as fp cartesian coords
    x = (x - y / 2) * scale + 400
    y = y * 1.732 / 2 * scale + 400
  End Sub

  Private Function CV(ByVal x)
    Return (x - 500) * scale + 400
  End Function

  Private Sub UpdateStats()
    Label5.Text = "Generations:" & Str(Ngens)
    Label6.Text = "Live cells:" & Str(Ncells)
  End Sub

  Private Sub Backward(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button1.Click
    Go(-100)
  End Sub

  Private Sub Backward10(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button2.Click
    Go(-10)
  End Sub

  Private Sub Backward1(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button3.Click
    Go(-1)
  End Sub

  Private Sub Halt(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button4.Click
    Go(0)
  End Sub

  Private Sub Forward1(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button5.Click
    Go(1)
  End Sub

  Private Sub Forward10(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button6.Click
    Go(10)
  End Sub

  Private Sub Forward(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button7.Click
    Go(100)
  End Sub

  Private Sub Clear(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button8.Click
    Clear2()
  End Sub

  Private Sub Clear2()
    For i As Integer = 0 To 1000
      For j As Integer = 0 To 1000
        Grid(i, j) = 0
      Next
    Next
    Ngens = 0 : Ncells = 0 : UpdateStats()
    Xmin = 480 : Xmax = 520 : Ymin = 480 : Ymax = 520
    ToDo = 0 : BufferPointer = 0
    If RadioButton5.Checked Then
      scale = AutoScale()
    Else
      RadioButton1.Checked = True
      RadioButton2.Checked = False
      RadioButton3.Checked = False
      RadioButton4.Checked = False
    End If
    PictureBox1.Invalidate()
  End Sub

  Private Sub Go(ByVal g)
    CheckRules()
    ToDo = g
  End Sub

  Private Function CheckRules() As Boolean
    BirthMin = NumericUpDown1.Value
    BirthMax = NumericUpDown2.Value
    SurviveMin = NumericUpDown3.Value
    SurviveMax = NumericUpDown4.Value
    If BirthMin > BirthMax Or SurviveMin > SurviveMax Then
      GroupBox3.BackColor = Color.MistyRose : Return False
    Else
      GroupBox3.BackColor = Color.FromArgb(248, 248, 248)
    End If
    Return True
  End Function

  Private Function CountNeighbours(ByVal i, ByVal j) As Integer
    ' NB this includes the cell itself
    Dim n As Integer = 0
    Select Case mode
      Case "Hexagonal"
        If Grid(i - 1, j - 1) And 2 Then n += 1
        If Grid(i, j - 1) And 2 Then n += 1
        If Grid(i - 1, j) And 2 Then n += 1
        If Grid(i, j) And 2 Then n += 1
        If Grid(i + 1, j) And 2 Then n += 1
        If Grid(i, j + 1) And 2 Then n += 1
        If Grid(i + 1, j + 1) And 2 Then n += 1
      Case "Rectangular"
        For p As Integer = i - 1 To i + 1
          For q As Integer = j - 1 To j + 1
            If Grid(p, q) And 2 Then n += 1
          Next
        Next
    End Select
    Return n
  End Function

  Private Sub ScaleChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles _
  RadioButton1.CheckedChanged, RadioButton2.CheckedChanged, RadioButton2.CheckedChanged, _
  RadioButton4.CheckedChanged, RadioButton5.CheckedChanged
    Dim s As Integer
    Select Case True
      Case RadioButton1.Checked : s = 20
      Case RadioButton2.Checked : s = 10
      Case RadioButton3.Checked : s = 4
      Case RadioButton4.Checked : s = 1
      Case RadioButton5.Checked : s = AutoScale()
    End Select
    If scale <> s Then
      scale = s
      PictureBox1.Invalidate()
    End If
  End Sub

  Private Shadows Function AutoScale()
    Dim s1, s2 As Integer
    s1 = 500 - Xmin
    s2 = Xmax - 500 : If s2 > s1 Then s1 = s2
    s2 = 500 - Ymin : If s2 > s1 Then s1 = s2
    s2 = Ymax - 500 : If s2 > s1 Then s1 = s2
    Return 400 \ s1
  End Function

  Private Sub SpeedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles _
  RadioButton6.CheckedChanged, RadioButton7.CheckedChanged, RadioButton8.CheckedChanged
    Select Case True
      Case RadioButton6.Checked : speed = "slow"
      Case RadioButton7.Checked : speed = "moderate"
      Case RadioButton8.Checked : speed = "fast"
    End Select
  End Sub

  Private Sub ModeChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles _
  RadioButton9.CheckedChanged, RadioButton10.CheckedChanged
    Select Case True
      Case RadioButton9.Checked : mode = "Rectangular"
      Case RadioButton10.Checked : mode = "Hexagonal"
    End Select
    Clear2()
  End Sub

  Private Sub RulesChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles _
  NumericUpDown1.ValueChanged, NumericUpDown2.ValueChanged, NumericUpDown3.ValueChanged, NumericUpDown4.ValueChanged
    CheckRules()
  End Sub

  Private Sub BackgroundWorker1_DoWork(ByVal sender As Object, ByVal e As System.ComponentModel.DoWorkEventArgs) Handles BackgroundWorker1.DoWork
    Dim n As Integer
    Dim changes As Boolean
    Do
      If ToDo > 0 Then ' go forward
        CheckRules()
        For i As Integer = Xmin - 1 To Xmax + 1
          For j As Integer = Ymin - 1 To Ymax + 1
            If Grid(i, j) And 2 Then ' live cell
              n = CountNeighbours(i, j) - 1 ' don't count me
              If n >= SurviveMin And n <= SurviveMax Then Grid(i, j) = Grid(i, j) Or 1
            Else ' blank cell
              n = CountNeighbours(i, j)
              If n >= BirthMin And n <= BirthMax Then Grid(i, j) = Grid(i, j) Or 1
            End If
          Next
        Next
        Ncells = 0 : changes = False
        For i As Integer = Xmin - 1 To Xmax + 1
          For j As Integer = Ymin - 1 To Ymax + 1
            If (Grid(i, j) And 3) = 2 Or (Grid(i, j) And 3) = 1 Then
              ' cell is being changed
              BufferPointer += 1 : changes = True
              If BufferPointer > BufferSize Then BufferPointer = 1
              Buffer(BufferPointer) = i + (j << 12) + ((Grid(i, j) And 2) << 24)
              'Debug.Print(Hex(Buffer(BufferPointer)))
            End If
            If Grid(i, j) And &HD Then Grid(i, j) = Grid(i, j) Or 2
            Grid(i, j) = Grid(i, j) << 1
            If Grid(i, j) And 2 Then
              Ncells += 1
              UpdateLimits(i, j)
            End If
          Next
        Next
        BufferPointer += 1 : Buffer(BufferPointer) = Ncells Or (1 << 31) ' end marker
        Ngens += 1 : UpdateStats()
        If RadioButton5.Checked Then scale = AutoScale()
        PictureBox1.Invalidate()
        Application.DoEvents()
        If ToDo < 100 Then ToDo -= 1
        If Ncells = 0 Or changes = False Then ToDo = 0
        If ToDo > 0 Then
          Select Case speed
            Case "slow" : Wait(1000)
            Case "moderate" : Wait(100)
            Case "fast"
          End Select
        End If
      End If
      If ToDo < 0 Then ' go backward
        Dim i, j, c As Integer
        If BufferPointer > 0 Then BufferPointer -= 1 ' move back over end marker
        While BufferPointer > 0 And Buffer(BufferPointer) > 0
          i = Buffer(BufferPointer) And &HFFF
          j = (Buffer(BufferPointer) >> 12) And &HFFF
          c = (Buffer(BufferPointer) >> 24) And &HFFF
          Grid(i, j) = (c And 2)
          BufferPointer -= 1
        End While
        Ncells = Buffer(BufferPointer) And &HFFFFFF
        Ngens -= 1 : UpdateStats()
        PictureBox1.Invalidate()
        Application.DoEvents()
        If ToDo > -100 Then ToDo += 1
        If ToDo < -1 Then
          Select Case speed
            Case "slow" : Wait(1000)
            Case "moderate" : Wait(100)
            Case "fast"
          End Select
        End If
      End If
    Loop
  End Sub

  Private Sub Wait(ByVal n)
    Dim t As Integer = My.Computer.Clock.TickCount
    Do
    Loop Until My.Computer.Clock.TickCount > t + n
  End Sub

  Private Sub EditList(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button9.Click
    If EditListDialog.ShowDialog() = Windows.Forms.DialogResult.OK Then
      ComboBox1.Items.Clear()
      For i As Integer = 0 To Nlist - 1
        ComboBox1.Items.Add(GetName(List(i)))
      Next
      SaveList()
    End If
  End Sub
End Class
